(* static/statobj.sml *)

functor StatObjFn(structure Absyn : ABSYN
		  structure TyFcn : TY_FCN
		  structure TyScheme : TY_SCHEME
		  sharing TyFcn.Ty = TyScheme.Ty
		    ) : STAT_OBJ =
  struct

    structure Absyn	= Absyn
    structure IdentDict = Absyn.IdentDict
    structure Ty	= TyFcn.Ty
    structure TyFcn	= TyFcn
    structure TyScheme	= TyScheme

    datatype valkind	= CON | VAR | REL
    datatype valstr	= VALSTR of {vk: valkind,
				     sigma: TyScheme.tyscheme}
    datatype tystr	= TYSTR of {theta: TyFcn.tyfcn,
				    abstract: bool}
    datatype modstr	= MODSTR of {TE: tystr IdentDict.dict,
				     VE: valstr IdentDict.dict,
				     source: Absyn.Source.source}

    (* standard type names *)
    val t_bool	= Ty.TYNAME{modid="RML",tycon="bool",eq=ref Ty.MAYBE}
    val t_char	= Ty.TYNAME{modid="RML",tycon="char",eq=ref Ty.MAYBE}
    val t_int	= Ty.TYNAME{modid="RML",tycon="int",eq=ref Ty.MAYBE}
    val t_list	= Ty.TYNAME{modid="RML",tycon="list",eq=ref Ty.MAYBE}
    val t_lvar	= Ty.TYNAME{modid="RML",tycon="lvar",eq=ref Ty.ALWAYS}
    val t_option= Ty.TYNAME{modid="RML",tycon="option",eq=ref Ty.MAYBE}
    val t_real	= Ty.TYNAME{modid="RML",tycon="real",eq=ref Ty.MAYBE}
    val t_string= Ty.TYNAME{modid="RML",tycon="string",eq=ref Ty.MAYBE}
    val t_vector= Ty.TYNAME{modid="RML",tycon="vector",eq=ref Ty.MAYBE}

    (* standard atomic types *)
    val tau_bool	= Ty.CONS([],t_bool)
    val tau_char	= Ty.CONS([],t_char)
    val tau_int		= Ty.CONS([],t_int)
    val tau_real	= Ty.CONS([],t_real)
    val tau_string	= Ty.CONS([],t_string)

    (* some standard types *)
    val alpha		= Ty.RIGID "a"
    val tau_alpha	= Ty.VAR alpha
    val tau_alpha_list	= Ty.CONS([tau_alpha], t_list)
    val tau_alpha_lvar	= Ty.CONS([tau_alpha], t_lvar)
    val tau_alpha_option= Ty.CONS([tau_alpha], t_option)
    val tau_alpha_vector= Ty.CONS([tau_alpha], t_vector)
    val tau_char_list	= Ty.CONS([tau_char], t_list)
    val eqalpha		= Ty.RIGID "'a"
    val tau_eqalpha	= Ty.VAR eqalpha
    val tau_eqalpha_list= Ty.CONS([tau_eqalpha], t_list)

    (* Source object for the standard environment *)

    val sourceInit =
      Absyn.Source.SOURCE{fileName = "(pervasive RML environment)",
			  newLines = []}

    (* standard value environment *)
    val VE_init =
      let fun bind(VE, id, tau, vk) =
	    let val sigma = TyScheme.genAll tau
		val id = Absyn.rmlIdent id
	    in
	      IdentDict.insert(VE, id, VALSTR{vk=vk, sigma=sigma})
	    end
	  fun bindRel(VE, var, domtaus, codtaus) =
	    bind(VE, var, Ty.REL(domtaus,codtaus), REL)
	  fun bindCon(VE, con, tau) = bind(VE, con, tau, CON)
	  val VE = IdentDict.empty
	  (* options *)
	  val VE = bindCon(VE,"NONE",tau_alpha_option)
	  val VE = bindCon(VE,"SOME",Ty.REL([tau_alpha],[tau_alpha_option]))
	  (* booleans *)
	  val VE = bindCon(VE,"false",tau_bool)
	  val VE = bindCon(VE,"true",tau_bool)
	  val VE = bindRel(VE,"bool_and",[tau_bool,tau_bool],[tau_bool])
	  val VE = bindRel(VE,"bool_not",[tau_bool],[tau_bool])
	  val VE = bindRel(VE,"bool_or",[tau_bool,tau_bool],[tau_bool])
	  (* characters *)
	  val VE = bindRel(VE,"char_int",[tau_char],[tau_int])
	  (* integers *)
	  val VE = bindRel(VE,"int_abs",[tau_int],[tau_int])
	  val VE = bindRel(VE,"int_add",[tau_int,tau_int],[tau_int])
	  val VE = bindRel(VE,"int_char",[tau_int],[tau_char])
	  val VE = bindRel(VE,"int_div",[tau_int,tau_int],[tau_int])
	  val VE = bindRel(VE,"int_eq",[tau_int,tau_int],[tau_bool])
	  val VE = bindRel(VE,"int_ge",[tau_int,tau_int],[tau_bool])
	  val VE = bindRel(VE,"int_gt",[tau_int,tau_int],[tau_bool])
	  val VE = bindRel(VE,"int_le",[tau_int,tau_int],[tau_bool])
	  val VE = bindRel(VE,"int_lt",[tau_int,tau_int],[tau_bool])
	  val VE = bindRel(VE,"int_max",[tau_int,tau_int],[tau_int])
	  val VE = bindRel(VE,"int_min",[tau_int,tau_int],[tau_int])
	  val VE = bindRel(VE,"int_mod",[tau_int,tau_int],[tau_int])
	  val VE = bindRel(VE,"int_mul",[tau_int,tau_int],[tau_int])
	  val VE = bindRel(VE,"int_ne",[tau_int,tau_int],[tau_bool])
	  val VE = bindRel(VE,"int_neg",[tau_int],[tau_int])
	  val VE = bindRel(VE,"int_real",[tau_int],[tau_real])
	  val VE = bindRel(VE,"int_string",[tau_int],[tau_string])
	  val VE = bindRel(VE,"int_sub",[tau_int,tau_int],[tau_int])
	  (* lists *)
	  val VE = bindCon(VE,"cons",Ty.REL([tau_alpha,tau_alpha_list],[tau_alpha_list]))
	  val VE = bindCon(VE,"nil",tau_alpha_list)
	  val VE = bindRel(VE,"list_append",[tau_alpha_list,tau_alpha_list],[tau_alpha_list])
	  val VE = bindRel(VE,"list_delete",[tau_alpha_list,tau_int],[tau_alpha_list])
	  val VE = bindRel(VE,"list_length",[tau_alpha_list],[tau_int])
	  val VE = bindRel(VE,"list_member",[tau_eqalpha,tau_eqalpha_list],[tau_bool])
	  val VE = bindRel(VE,"list_nth",[tau_alpha_list,tau_int],[tau_alpha])
	  val VE = bindRel(VE,"list_reverse",[tau_alpha_list],[tau_alpha_list])
	  val VE = bindRel(VE,"list_string",[tau_char_list],[tau_string])
	  val VE = bindRel(VE,"list_vector",[tau_alpha_list],[tau_alpha_vector])
	  (* logical variables *)
	  val VE = bindRel(VE,"lvar_get",[tau_alpha_lvar],[tau_alpha_option])
	  val VE = bindRel(VE,"lvar_new",[],[tau_alpha_lvar])
	  val VE = bindRel(VE,"lvar_set",[tau_alpha_lvar,tau_alpha],[])
	  (* reals *)
	  val VE = bindRel(VE,"real_abs",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_add",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_atan",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_cos",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_div",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_eq",[tau_real,tau_real],[tau_bool])
	  val VE = bindRel(VE,"real_exp",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_floor",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_ge",[tau_real,tau_real],[tau_bool])
	  val VE = bindRel(VE,"real_gt",[tau_real,tau_real],[tau_bool])
	  val VE = bindRel(VE,"real_int",[tau_real],[tau_int])
	  val VE = bindRel(VE,"real_le",[tau_real,tau_real],[tau_bool])
	  val VE = bindRel(VE,"real_ln",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_lt",[tau_real,tau_real],[tau_bool])
	  val VE = bindRel(VE,"real_max",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_min",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_mod",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_mul",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_ne",[tau_real,tau_real],[tau_bool])
	  val VE = bindRel(VE,"real_neg",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_pow",[tau_real,tau_real],[tau_real])
	  val VE = bindRel(VE,"real_sin",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_sqrt",[tau_real],[tau_real])
	  val VE = bindRel(VE,"real_string",[tau_real],[tau_string])
	  val VE = bindRel(VE,"real_sub",[tau_real,tau_real],[tau_real])
	  (* strings *)
	  val VE = bindRel(VE,"string_append",[tau_string,tau_string],[tau_string])
	  val VE = bindRel(VE,"string_int",[tau_string],[tau_int])
	  val VE = bindRel(VE,"string_length",[tau_string],[tau_int])
	  val VE = bindRel(VE,"string_list",[tau_string],[tau_char_list])
	  val VE = bindRel(VE,"string_nth",[tau_string,tau_int],[tau_char])
	  (* vectors *)
	  val VE = bindRel(VE,"vector_length",[tau_alpha_vector],[tau_int])
	  val VE = bindRel(VE,"vector_list",[tau_alpha_vector],[tau_alpha_list])
	  val VE = bindRel(VE,"vector_nth",[tau_alpha_vector,tau_int],[tau_alpha])
	  (* misc *)
	  val VE = bindRel(VE,"clock",[],[tau_real])
	  val VE = bindRel(VE,"print",[tau_string],[])
	  val VE = bindRel(VE,"tick",[],[tau_int])
      in
	VE
      end

    (* standard type environment *)
    val TE_init =
      let fun bind(TE, tycon, theta) =
	    let val tystr = TYSTR{theta=theta, abstract=false}
		val tycon = Absyn.rmlIdent tycon
	    in
	      IdentDict.insert(TE, tycon, tystr)
	    end
	  val TE = IdentDict.empty
	  val TE = bind(TE,"bool",TyFcn.lambda([],tau_bool))
	  val TE = bind(TE,"char",TyFcn.lambda([],tau_char))
	  val TE = bind(TE,"int",TyFcn.lambda([],tau_int))
	  val TE = bind(TE,"list",TyFcn.lambda([alpha],tau_alpha_list))
	  val TE = bind(TE,"lvar",TyFcn.lambda([alpha],tau_alpha_lvar))
	  val TE = bind(TE,"option",TyFcn.lambda([alpha],tau_alpha_option))
	  val TE = bind(TE,"real",TyFcn.lambda([],tau_real))
	  val TE = bind(TE,"string",TyFcn.lambda([],tau_string))
	  val TE = bind(TE,"vector",TyFcn.lambda([alpha],tau_alpha_vector))
      in
	TE
      end

    (* standard module environment *)
    val ME_init =
      let val modstr = MODSTR{TE=TE_init, VE=VE_init, source=sourceInit}
	  val modid = Absyn.rmlIdent "RML"
      in
	IdentDict.insert(IdentDict.empty, modid, modstr)
      end

  end (* functor StatObjFn *)
